from ssl import CertificateError, _inet_paton, _dnsname_match, _ipaddress_match

def loosely_dnsname_match(dn, hostname):
    return dn.rsplit('.', maxsplit=2)[-2:] == hostname.rsplit('.', maxsplit=2)[-2:]

def match_hostname(cert, hostname, policy):
    """based on ssl.py in python3.11
    """
    try:
        host_ip = _inet_paton(hostname)
    except ValueError:
        # Not an IP address (common case)
        host_ip = None
    dnsnames = []
    san = cert.get('subjectAltName', ())
    for key, value in san:
        if key == 'DNS':
            if host_ip is None and (_dnsname_match(value, hostname) or (policy != 'strict' and loosely_dnsname_match(value, hostname))):
                return True
            dnsnames.append(value)
        elif key == 'IP Address':
            if host_ip is not None and _ipaddress_match(value, host_ip):
                return True
            dnsnames.append(value)
    if not dnsnames:
        # The subject is only checked when there is no dNSName entry
        # in subjectAltName
        for sub in cert.get('subject', ()):
            for key, value in sub:
                # XXX according to RFC 2818, the most specific Common Name
                # must be used.
                if key == 'commonName':
                    if _dnsname_match(value, hostname) or (policy != 'strict' and loosely_dnsname_match(value, hostname)):
                        return False
                    dnsnames.append(value)
    if len(dnsnames) > 1:
        return False
    elif len(dnsnames) == 1:
        return False
    else:
        raise CertificateError("no appropriate commonName or "
            "subjectAltName fields were found")


import sys
from . import setting

if sys.platform.startswith('win'):
    from ctypes import windll
    import winreg

#TODO: use Windows API instead of winreg, see https://github.com/Noisyfox/sysproxy for example.

def set_pac(pac):
    if setting.config['setproxy'] and sys.platform.startswith('win'):
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER,
                                        r'Software\Microsoft\Windows\CurrentVersion\Internet Settings',
                                        0, winreg.KEY_ALL_ACCESS) as INTERNET_SETTINGS:
            if pac is None:
                winreg.DeleteValue(INTERNET_SETTINGS, 'AutoConfigURL')
            else:
                winreg.SetValueEx(INTERNET_SETTINGS, 'AutoConfigURL', 0, winreg.REG_SZ, pac)

        INTERNET_OPTION_REFRESH = 37
        INTERNET_OPTION_SETTINGS_CHANGED = 39
        windll.Wininet.InternetSetOptionW(0, INTERNET_OPTION_REFRESH, 0, 0)
        windll.Wininet.InternetSetOptionW(0, INTERNET_OPTION_SETTINGS_CHANGED, 0, 0)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Accesser
# Copyright (C) 2018  URenko

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa, ec
from cryptography.x509.extensions import (
    AuthorityKeyIdentifier,
    KeyUsage,
    ExtendedKeyUsage,
    SubjectKeyIdentifier,
)
from cryptography.x509.oid import ExtendedKeyUsageOID

from .log import logger

logger = logger.getChild("certmanager")
from . import setting


def create_root_ca():
    key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=4096,
    )

    subject = issuer = x509.Name(
        [
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Accesser"),
            x509.NameAttribute(NameOID.COMMON_NAME, "Accesser"),
        ]
    )
    cert = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer)
        .public_key(key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.datetime.now(datetime.timezone.utc))
        .not_valid_after(
            datetime.datetime.now(datetime.timezone.utc)
            + datetime.timedelta(days=25 * 365)
        )
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=None),
            critical=True,
        )
        .add_extension(
            KeyUsage(
                digital_signature=False,
                content_commitment=False,
                key_encipherment=False,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=True,
                crl_sign=True,
                encipher_only=False,
                decipher_only=False,
            ),
            critical=True,
        )
        .add_extension(
            SubjectKeyIdentifier.from_public_key(key.public_key()),
            critical=False,
        )
        .sign(key, hashes.SHA256())
    )

    setting.certpath.joinpath("root.crt").write_bytes(
        cert.public_bytes(serialization.Encoding.PEM)
    )

    setting.certpath.joinpath("root.key").write_bytes(
        key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
    )

    setting.certpath.joinpath("root.pfx").write_bytes(
        serialization.pkcs12.serialize_key_and_certificates(
            b"Accesser", key, cert, None, serialization.NoEncryption()
        )
    )

pkey = ec.generate_private_key(ec.SECP256R1())

def create_certificate(server_name):
    rootpem = setting.certpath.joinpath("root.crt").read_bytes()
    rootkey = setting.certpath.joinpath("root.key").read_bytes()
    ca_cert = x509.load_pem_x509_certificate(rootpem)
    ca_key = serialization.load_pem_private_key(rootkey, password=None)

    cert = (
        x509.CertificateBuilder()
        .subject_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.ORGANIZATION_NAME, "Accesser"),
                    x509.NameAttribute(NameOID.COMMON_NAME, "Accesser_Proxy"),
                ]
            )
        )
        .issuer_name(ca_cert.subject)
        .public_key(pkey.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(
            datetime.datetime.now(datetime.timezone.utc)
            - datetime.timedelta(seconds=600)
        )
        .not_valid_after(
            datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=30)
        )
        .add_extension(
            x509.SubjectAlternativeName(
                [
                    x509.DNSName(server_name),
                    x509.DNSName("*." + server_name),
                ]
            ),
            critical=False,
        )
        .add_extension(
            KeyUsage(
                digital_signature=True,
                content_commitment=False,
                key_encipherment=False,
                data_encipherment=False,
                key_agreement=False,
                key_cert_sign=False,
                crl_sign=False,
                encipher_only=False,
                decipher_only=False,
            ),
            critical=True,
        )
        .add_extension(
            AuthorityKeyIdentifier.from_issuer_public_key(ca_key.public_key()),
            critical=False,
        )
        .add_extension(
            SubjectKeyIdentifier.from_public_key(pkey.public_key()),
            critical=False,
        )
        .add_extension(
            ExtendedKeyUsage(
                usages=[
                    ExtendedKeyUsageOID.SERVER_AUTH,
                    ExtendedKeyUsageOID.CLIENT_AUTH,
                ]
            ),
            critical=True,
        )
        .sign(ca_key, hashes.SHA256())
    )

    setting.certpath.joinpath(f"{server_name}.crt").write_bytes(
        cert.public_bytes(serialization.Encoding.PEM)
        + pkey.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
    )

    return cert



import logging, logging.handlers
try:
    from rich.logging import RichHandler
    has_rich = True
except ModuleNotFoundError:
    has_rich = False

from . import setting

is_console = not setting.config['log']['logfile']
logging.basicConfig(
    format="%(message)s" if has_rich and is_console else '%(asctime)s %(levelname)-8s %(name)s: %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S',
    handlers=[logging.handlers.RotatingFileHandler(setting.config['log']['logfile'], maxBytes=2**20, backupCount=1)] if not is_console else ([RichHandler()] if has_rich else None)
)
logger = logging.getLogger('Accesser')
logger.setLevel(setting.config['log']['loglevel'].upper())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Accesser
# Copyright (C) 2018  URenko

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__version__ = '0.11.1'

import os, sys
import datetime
import json
import fnmatch
import random
import ssl
import asyncio
import traceback
from urllib import request
from urllib.parse import urlsplit
from packaging.version import Version
from tld import get_tld, is_tld
import dns, dns.asyncresolver, dns.nameserver
import platform
from pathlib import Path

from .utils import certmanager as cm
from .utils import importca
from .utils import setting
from .utils.setting import basepath
from .utils.log import logger
from .utils.cert_verify import match_hostname
from .utils import sysproxy

domain_ssl_map : dict[str, tuple[ssl.SSLContext, datetime.datetime]]  # TODO: LRU cache, i.e. maximum length for dict

async def update_cert(server_name):
    global domain_ssl_map, cert_lock
    if not is_tld(server_name):
        res = get_tld(server_name, as_object=True, fix_protocol=True)
        if res.subdomain:
            server_name = f"{res.subdomain.split('.', 1)[-1]}.{res.domain}.{res.tld}"
        else:
            server_name = f'{res.domain}.{res.tld}'
    async with cert_lock:
        if not ((server_name in domain_ssl_map) and ((datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=1)) < domain_ssl_map[server_name][1])):
            cert = cm.create_certificate(server_name)
            context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
            context.load_cert_chain(setting.certpath.joinpath("{}.crt".format(server_name)))
            domain_ssl_map[server_name] = (context, cert.not_valid_after_utc)
            logger.debug("The certificate for %s is refreshed (expiry: %s).", server_name, cert.not_valid_after_utc)
        else:
            context = domain_ssl_map[server_name][0]
        return context

async def send_pac(writer: asyncio.StreamWriter):
    pac_file = Path('pac') if Path('pac').exists() else Path(basepath).joinpath('pac')
    pac = pac_file.read_bytes().replace(b'{{port}}', str(setting.config['server']['port']).encode('iso-8859-1')).replace(b'{{host}}', setting.config['server'].get('pac_host', '127.0.0.1').encode('iso-8859-1'))
    writer.write(f'HTTP/1.1 200 OK\r\nContent-Type: application/x-ns-proxy-autoconfig\r\nContent-Length: {len(pac)}\r\n\r\n'.encode('iso-8859-1'))
    writer.write(pac)
    await writer.drain()
    writer.close()
    await writer.wait_closed()

async def send_crt(writer: asyncio.StreamWriter, path: str):
    crt = setting.certpath.joinpath(path.rsplit(sep = '/',maxsplit = 1)[-1]).read_bytes()
    writer.write(f'HTTP/1.1 200 OK\r\nContent-Type: application/x-x509-ca-cert\r\nContent-Length: {len(crt)}\r\n\r\n'.encode('iso-8859-1'))
    writer.write(crt)
    await writer.drain()
    writer.close()
    await writer.wait_closed()
    
async def http_redirect(writer: asyncio.StreamWriter, path: str):
    path = path.removeprefix('http://')
    for key in setting.config['http_redirect']:
        if path.startswith(key):
            path = setting.config['http_redirect'][key] + path[len(key):]
            break
    logger.debug('Redirect to %s', path)
    writer.write(f'HTTP/1.1 301 Moved Permanently\r\nLocation: https://{path}\r\n\r\n'.encode('iso-8859-1'))
    await writer.drain()
    writer.close()
    await writer.wait_closed()

async def forward_stream(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
    while True:
        data = await reader.read(32768)
        if data == b'':
            return
        writer.write(data)
        await writer.drain()

async def handle(reader, writer):
    try:
        remote_writer = None
        raw_request = await reader.readuntil(b'\r\n\r\n')
        requestline = raw_request.decode('iso-8859-1').splitlines()[0]
        i_addr, i_port, *_ = writer.get_extra_info('peername')
        logger.debug("%s:%d say: %s", i_addr, i_port, requestline)
        words = requestline.split()
        command, path = words[:2]
        match command:
            case 'CONNECT':
                host, port = path.split(':')
                remote_ip = await DNSquery(host)
                logger.debug('[%5d] DNS: %s -> %s', i_port, host, remote_ip)
            case 'GET':
                if path.startswith('/pac/'):
                    return await send_pac(writer)
                elif path.startswith('/CERT/root.'):
                    return await send_crt(writer, path)
                elif path == '/shutdown':
                    sys.exit()
                else:
                    return await http_redirect(writer, path)
            case _:
                return await http_redirect(writer, path)
        writer.write(b'HTTP/1.1 200 Connection Established\r\n\r\n')

        context = await update_cert(host)
        if sys.version_info[1] >= 11:
            await writer.start_tls(context)
        else:
            writer._transport = await writer._loop.start_tls(writer.transport, writer._protocol, context, server_side=True)
        server_hostname_key = next(filter(lambda h:fnmatch.fnmatchcase(host, h), setting.config['alter_hostname']), None)
        server_hostname = '' if server_hostname_key is None else setting.config['alter_hostname'][server_hostname_key]
        logger.debug("[%5d] server_hostname: %s", i_port, server_hostname)
        remote_context = ssl.create_default_context()
        remote_context.check_hostname = False
        remote_reader, remote_writer = await asyncio.open_connection(remote_ip, port, ssl=remote_context, server_hostname=server_hostname)
        cert = remote_writer.get_extra_info('peercert')
        cert_message = f"subjectAltName: {cert.get('subjectAltName', ())}, subject: {cert.get('subject', ())}"
        logger.debug("[%5d] %s", i_port, cert_message)
        cert_verify_key = next(filter(lambda h:fnmatch.fnmatchcase(host, h), setting.config.get('cert_verify', ())), None)
        if cert_verify_key is not None:
            cert_verify_list = setting.config['cert_verify'][cert_verify_key]
            cert_policy = False if cert_verify_list is False else True
        elif server_hostname_key is not None:
            cert_verify_list = [setting.config['alter_hostname'][server_hostname_key]]
            cert_policy = setting.config['check_hostname']
        else:
            cert_verify_list = [host]
            cert_policy = setting.config['check_hostname']
        if cert_policy is not False and not any(match_hostname(cert, h, cert_policy) for h in cert_verify_list):
            logger.warning("[%5d] %s don't match either of %s.", i_port, cert_verify_list, cert_message)
            return
        tasks = (
            asyncio.create_task(forward_stream(reader, remote_writer)),
            asyncio.create_task(forward_stream(remote_reader, writer))
        )
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        await asyncio.gather(*pending, return_exceptions=True)
    finally:
        writer.close()
        if remote_writer:
            remote_writer.close()
            await remote_writer.wait_closed()
        await writer.wait_closed()

async def proxy():
    server = await asyncio.start_server(handle, setting.config['server']['address'], setting.config['server']['port'])

    print(f"Serving on {', '.join(str(sock.getsockname()) for sock in server.sockets)}")
    sysproxy.set_pac(f"http://localhost:{setting.config['server']['port']}/pac/?t={random.randrange(2**16)}")

    try:
        async with server:
            await server.serve_forever()
    finally:
        sysproxy.set_pac(None)

async def DNSquery(domain, hosts_only=False):
    global DNSresolver
    try:
        return next(v for k, v in setting.config['hosts'].items() if k == domain or (k.startswith('.') and domain.endswith(k)))
    except StopIteration:
        if hosts_only:
            return
    if setting.config['ipv6']:
        try:
            ret = await DNSresolver.resolve(domain, 'AAAA')
        except dns.asyncresolver.NoAnswer:
            ret = await DNSresolver.resolve(domain, 'A')
    else:
        ret = await DNSresolver.resolve(domain, 'A')
    return ret[0].to_text()

def update_checker():
    for pypi_url in ('https://pypi.org/pypi/accesser/json', 'https://mirrors.cloud.tencent.com/pypi/json/accesser'):
        try:
            with request.urlopen(pypi_url) as f:
                v2 = Version(json.load(f)["info"]["version"])
                break
        except Exception:
            logger.warning(traceback.format_exc())
    else:
        with request.urlopen('https://github.com/URenko/Accesser/releases/latest') as f:
            v2 = Version(f.geturl().rsplit('/', maxsplit=1)[-1])
    v1 = Version(__version__)
    if v2 > v1:
        logger.warning("There is a new version. You can update with `python3 -m pip install -U accesser` or download from GitHub.")

async def main():
    global domain_ssl_map, cert_lock, DNSresolver
    print(f"Accesser v{__version__}  Copyright (C) 2018-2024  URenko")
    setting.parse_args()
    if platform.system() == "Linux" or platform.system() == "FreeBSD":
        if os.geteuid() == 0:
            logger.warning(
                "Running Accesser as the root user carries certain risks. Do not use it in production."
            )
    if setting.rules_update_case in ('old', 'missing'):
        logger.warning("Updated rules.toml because it is %s.", setting.rules_update_case)
    elif setting.rules_update_case == 'modified':
        logger.warning("You've already modified rules.toml so it won't be updated automatically!")
    else:
        logger.debug("rules.toml status: %s", setting.rules_update_case)

    if any(_keys in setting._config for _keys in setting._rules):
        logger.warning("Some sections of config.toml overlap with rules.toml, config.toml has higher priority, but this may make rule updates ineffective.")
        
    DNSresolver = dns.asyncresolver.Resolver(configure=False)
    DNSresolver.cache = dns.resolver.LRUCache()
    for nameserver in setting.config['DNS']['nameserver']:
        if (_url := urlsplit(nameserver)).netloc == '':
            _url = urlsplit('//' + nameserver)
        address = await DNSquery(_url.hostname, hosts_only=True)
        match _url.scheme:
            case '':
                DNSresolver.nameservers.append(dns.nameserver.Do53Nameserver(_url.hostname if address is None else address, 53 if _url.port is None else _url.port))
            case 'https':
                DNSresolver.nameservers.append(dns.nameserver.DoHNameserver(nameserver, bootstrap_address=address))
            case 'tls':
                DNSresolver.nameservers.append(dns.nameserver.DoTNameserver(_url.hostname if address is None else address, 853 if _url.port is None else _url.port, hostname=_url.hostname))
            case 'quic':
                DNSresolver.nameservers.append(dns.nameserver.DoQNameserver(_url.hostname if address is None else address, 853 if _url.port is None else _url.port, server_hostname=_url.hostname))
    
    importca.import_ca()

    domain_ssl_map = {}
    cert_lock = asyncio.Lock()
    
    await asyncio.gather(
        asyncio.to_thread(update_checker),
        proxy()
    )

def run():
    asyncio.run(main())



from pathlib import Path
import shutil, os, shutil, filecmp

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib
import argparse
import platform
import logging

basepath = Path(__file__).parent.parent
certpath = None


def deep_merge(config_a: dict, config_b: dict):
    res = config_b | config_a
    for key in config_a:
        if key in config_b and type(config_a[key]) is type(config_b[key]):
            if isinstance(config_a[key], dict):
                res[key] = deep_merge(config_a[key], config_b[key])
            elif isinstance(config_a[key], list):
                for elem in config_b[key]:
                    if elem not in res[key]:
                        res[key].append(elem)
    return res


if Path("rules.toml").exists():
    if Path("rules.toml").stat().st_mtime_ns > 0:
        rules_update_case = "modified"
    else:
        if filecmp.cmp("rules.toml", basepath / "rules.toml"):
            rules_update_case = "holding"
        else:
            rules_update_case = "old"
else:
    rules_update_case = "missing"
if rules_update_case in ("old", "missing"):
    shutil.copyfile(basepath / "rules.toml", "rules.toml")
    os.utime("rules.toml", ns=(0, 0))

_rules = {}

if not Path("rules").exists() or not Path("rules").is_dir():
    Path("rules").mkdir()

for custom_rules in sorted(Path("rules").glob("*.toml"), key=lambda f: f.name):
    with custom_rules.open(mode="rb") as f:
        _rules = deep_merge(_rules, tomllib.load(f))

with Path("rules.toml").open(mode="rb") as f:
    _rules = deep_merge(_rules, tomllib.load(f))

if not Path("config.toml").exists():
    shutil.copyfile(basepath / "config.toml", "config.toml")

with open("config.toml", "rb") as f:
    _config = tomllib.load(f)

config = _rules.copy()
config = deep_merge(_config, _rules)


def decide_state_path_legacy():
    if config["importca"]:
        return Path(basepath)
    return Path()


def decide_state_path_unix_like():
    if os.geteuid() == 0:
        return Path("/var/lib") / "accesser"

    state_path = os.getenv("XDG_STATE_HOME", None)
    if state_path is not None:
        state_path = Path(state_path) / "accesser"
    else:
        state_path = Path.home() / ".local/state" / "accesser"
    return state_path


def decide_certpath():
    # 人为指定最优先
    if "state_dir" in config and config["state_dir"] is not None:
        return Path(config["state_dir"]) / "CERT"
    match platform.system():
        case "Linux" | "FreeBSD":
            deprecated_path = decide_state_path_legacy() / "CERT"
            if deprecated_path.exists():
                logging.warning("cert path %s is deprecated.", str(deprecated_path))
                logging.warning("Please check https://github.com/URenko/Accesser/pull/245 for migration.")
                return deprecated_path
            return decide_state_path_unix_like() / "CERT"
        case _:
            # windows,mac,android ...
            return decide_state_path_legacy() / "CERT"
    return


def parse_args():
    global certpath
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--notsetproxy",
        action="store_true",
        help="do not set system's pac proxy automatically",
    )
    parser.add_argument(
        "--notimportca",
        action="store_true",
        help="do not import certificate to system automatically",
    )
    parser.add_argument(
        "--state-dir",
        type=str,
        help="where state file store , override notimportca",
        default=None,
    )
    args = parser.parse_args()
    if args.notsetproxy:
        config["setproxy"] = False
    if args.notimportca:
        config["importca"] = False
    if args.state_dir is not None:
        config["state_dir"] = args.state_dir
    certpath = decide_certpath()
    if not certpath.exists() or certpath.is_file():
        certpath.mkdir(parents=True)
    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Accesser
# Copyright (C) 2018  URenko

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, sys, platform
import datetime
import subprocess
import locale
from pathlib import Path

from cryptography import x509
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.serialization import pkcs12

from . import setting
from . import certmanager as cm
from .log import logger
logger = logger.getChild('importca')


def run_and_log(cmd, check=True):
    result = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT, 
        text=True,
    )

    logger.debug(result.stdout)

    if check:
        result.check_returncode()

    return result


def export_ca() -> bool:
    if platform.system() == 'Windows':
        try:
            pfx_path = setting.certpath.joinpath("root.pfx")
            ps_cmd = f"""
            $env:PSModulePath = [Environment]::GetEnvironmentVariable('PSModulePath', 'Machine')
            $pwd = ConvertTo-SecureString -String "Accesser" -AsPlainText -Force
            $cert = Get-ChildItem Cert:\\CurrentUser\\My | Where-Object {{ $_.Subject -like "CN=Accesser*" }} |
                    Sort-Object NotBefore -Descending | Select-Object -First 1
            if ($null -eq $cert) {{ exit 1 }}
            Export-PfxCertificate -Cert $cert -FilePath "{pfx_path}" -Password $pwd -Force
            """
            run_and_log(["powershell", "-NoProfile", "-Command", ps_cmd], check=True)
            logger.debug("Export succeed.")
        except subprocess.CalledProcessError:
            logger.debug("Export failed.")
            return False
        private_key, certificate, _ = pkcs12.load_key_and_certificates(pfx_path.read_bytes(), password=b'Accesser')
        setting.certpath.joinpath("root.crt").write_bytes(certificate.public_bytes(serialization.Encoding.PEM))
        setting.certpath.joinpath("root.key").write_bytes(
            private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
        setting.certpath.joinpath("root.pfx").write_bytes(
            serialization.pkcs12.serialize_key_and_certificates(
                b"Accesser", private_key, certificate, None, serialization.NoEncryption()
            )
        )
        return True
    else: # TODO
        return False


def import_windows_ca():
    try:
        logger.info("Importing new certificate")
        run_and_log(f'CertUtil -f -user -p "" -importPFX My "{setting.certpath.joinpath("root.pfx")}"')
    except subprocess.CalledProcessError:
        logger.error("Import Failed")
        run_and_log('CertUtil -user -delstore My Accesser')
        logger.warning('Try to manually import the certificate')


def import_mac_ca():
    ca_hash = CertUtil.ca_thumbprint.replace(':', '')

    def get_exist_ca_sha1():
        args = ['security', 'find-certificate', '-Z', '-a', '-c', 'Accesser']
        output = subprocess.check_output(args)
        for line in output.splitlines(True):
            if len(line) == 53 and line.startswith("SHA hash:"):
                sha1_hash = line[12:52]
                return sha1_hash

    exist_ca_sha1 = get_exist_ca_sha1()
    if exist_ca_sha1 == ca_hash:
        logger.info("Accesser CA exist")
        return

    import_command = 'security add-trusted-cert -d -r trustRoot -k /Library/Keychains/System.keychain {}'.format('root.crt')
    if exist_ca_sha1:
        delete_ca_command = 'security delete-certificate -Z %s' % exist_ca_sha1
        exec_command = "%s;%s" % (delete_ca_command, import_command)
    else:
        exec_command = import_command

    admin_command = """osascript -e 'do shell script "%s" with administrator privileges' """ % exec_command
    cmd = admin_command.encode('utf-8')
    logger.info("try auto import CA command:%s", cmd)
    os.system(cmd)


def is_cert_valid(cert_path: Path, key_path: Path, ensure_trusted: bool) -> int:
    cert = x509.load_pem_x509_certificate(cert_path.read_bytes())

    if (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=365)) > cert.not_valid_after_utc:
        logger.warning('The root certificate has expired or will expire soon!')
        return False

    private_key = serialization.load_pem_private_key(key_path.read_bytes(), password=None)

    public_key_cert = cert.public_key()
    public_key_private = private_key.public_key()
    if public_key_cert.public_numbers() != public_key_private.public_numbers():
        logger.warning('The root certificate and private key do not match!')
        return False

    if ensure_trusted:
        if sys.platform.startswith("win"):
            cert_fingerprint = cert.fingerprint(hashes.SHA1()).hex().upper()

            ps_cmd = f"""
            $env:PSModulePath = [Environment]::GetEnvironmentVariable('PSModulePath', 'Machine')
            $found = $false
            Get-ChildItem Cert:\\CurrentUser\\Root, Cert:\\LocalMachine\\Root | ForEach-Object {{
                if ($_.Thumbprint -eq '{cert_fingerprint}') {{ $found = $true }}
            }}
            exit ($found -eq $false)
            """
            result = run_and_log(["powershell", "-NoProfile", "-Command", ps_cmd], check=False)
            if result.returncode != 0:
                logger.warning('The root certificate is not yet trusted by the operating system!')
                return False

    return True


def import_ca():
    cert_path = setting.certpath.joinpath("root.crt")
    key_path = setting.certpath.joinpath("root.key")

    if not (cert_path.exists() and key_path.exists()) or not is_cert_valid(cert_path, key_path, ensure_trusted=setting.config['importca']):
        if not (export_ca() and is_cert_valid(cert_path, key_path, ensure_trusted=setting.config['importca'])):
            logger.warning('Generating a new root certificate...')
            cm.create_root_ca()
            if setting.config['importca']:
                if platform.system() == 'Windows':
                    import_windows_ca()
                    return
                else:
                    logger.warning('Automatic import of root certificate root.crt is not yet supported on this platform.')
    _log_msg = f'You can GET root certificate from http://localhost:{setting.config["server"]["port"]}/CERT/root.crt'
    if not setting.config['importca']:
        _log_msg += ' and import it manually.'
    logger.warning(_log_msg)
